from .continuous import *
from .discrete import *
from .global_planning import *
from .goal_conditioned_wrappers import *
from .agent import Agent
from .agents_index import AgentsIndex
from .value_based_agent import ValueBasedAgent


separator = ","
file_name = "maze_small.txt"
output_file_name = file_name.replace("txt", "csv")

with open(output_file_name, "w") as output_file:
    with open(file_name) as maze_file:
        for line in maze_file:
            if line[0] == '#':
                continue
            output_file.write(separator.join(list(line)))




from typing import Union
import numpy as np
import torch

from hbrl.agents.agent import Agent
from abc import ABC, abstractmethod
from gym.spaces import Box, Discrete
from hbrl.agents.utils.replay_buffer import ReplayBuffer


class ValueBasedAgent(Agent, ABC):

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """
        super().__init__(state_space, action_space, **params)

        self.batch_size = params.get("batch_size", 100)
        self.buffer_max_size = params.get("buffer_max_size", int(1e6))
        self.replay_buffer = ReplayBuffer(self.buffer_max_size, self.device)

    @abstractmethod
    def get_value(self, features, actions=None):
        return 0

    @abstractmethod
    def learn(self):
        pass

    def save_interaction(self, *interaction_data):
        """
        Function that is called to ask our agent to learn about the given interaction. This function is separated from
        self.on_action_stop(**interaction_data) because we can imagine agents that do not learn on every interaction, or
        agents that learn on interaction they didn't make (like HER that add interaction related to fake goals in their
        last trajectory).
        on_action_stop is used to update variables likes self.last_state or self.simulation_time_step_id, and
        learn_interaction is used to know the set of interactions we can learn about.

        Example: Our implementation of HER show a call to 'learn_interaction' without 'on_action_stop'
        (two last lines of 'her' file).
        """
        assert not self.under_test
        self.replay_buffer.append(interaction_data)

    def sample_training_batch(self, batch_size=None):
        batch_size = self.batch_size if batch_size is None else batch_size
        return self.replay_buffer.sample(batch_size)

    def process_interaction(self, action, reward, new_state, done, learn=True):
        if learn and not self.under_test:
            self.save_interaction(self.last_state, action, reward, new_state, done)
            self.learn()
        super().process_interaction(action, reward, new_state, done, learn=learn)



import os
import shutil
import datetime
import cv2
import numpy as np
from PIL import Image


def create_dir(dir_name):
    if os.path.isdir(dir_name):
        return
    dir_parts = dir_name.split("/")
    directory_to_create = ""
    for part in dir_parts:
        directory_to_create += part + "/"
        if not os.path.isdir(directory_to_create):
            try:
                os.mkdir(directory_to_create)
            except FileNotFoundError:
                print("failed to create dir " + str(directory_to_create))
                raise Exception


def empty_dir(directory_path):
    for filename in os.listdir(directory_path):
        file_path = os.path.join(directory_path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


def generate_video(images, output_directory: str, filename, convert_to_bgr=True):
    """
    generate a video from the given list of images, and save them at location output_directory/filename.mp4.
    @param images: List of images as numpy array of pixels. For each image, the expected shape is width * height * 3.
        images[n][-1] is expected to be a list of rgb pixels. But BGR pixels are accepted if convert_to_bgr is set to
        false.
    @param output_directory: A path. A '/' is added at the end if there's none in the given path.
    @param filename: a filename. Should not contain "/" characters or '.' except for the extension. If no '.' is found
        (aka no extension) a ".mp4" is added at the end.
    @param convert_to_bgr: (boolean) If True (default value), the colors are considered as RGB and are converted to BGR
        (which is the default opencv standard, don't ask me why).
    """
    # Convert image colors
    if convert_to_bgr:
        images = [cv2.cvtColor(img.astype(np.uint8), cv2.COLOR_RGB2BGR) for img in images]

    if output_directory[-1] != "/":
        output_directory += "/"
    assert output_directory.find(".")

    # Verify filename
    if len(filename) < 4 or filename[-4:] != ".mp4":
        filename += ".mp4"
    assert len(filename.split(".")) == 2

    create_dir(output_directory)
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    height, width, channels = images[0].shape
    fps = 30

    out = cv2.VideoWriter(output_directory + filename, fourcc, fps, (width, height))
    for image_data in images:
        image_data = image_data.astype(np.uint8)
        out.write(image_data)
    out.release()


def save_image(image: np.ndarray, directory, file_name):
    if directory[-1] != "/":
        directory += "/"
    image = Image.fromarray(image)
    create_dir(directory)
    if not file_name.endswith(".png"):
        if len(file_name.split(".")) > 1:
            file_name = "".join(file_name.split(".")[:-1])  # Remove the last extension
        assert len(file_name.split(".")) == 1
        file_name += ".png"
    image.save(directory + file_name)


def get_red_green_color(value, hexadecimal=True):
    """
    Return a colour that belongs to a gradiant from red (value=0) to green (value=1).
    @param value: value between 0 and 1 that defines result color.
    @param hexadecimal: THe colour will be return in hexadecimal if true, in a list of RGB int otherwise.
    """
    low_color = [255, 0, 0]
    high_color = [0, 255, 0]
    if hexadecimal:
        result = "#"
    else:
        result = []
    for index, (low, high) in enumerate(zip(low_color, high_color)):
        difference = high - low
        if hexadecimal:
            final_color = hex(int(low + value * difference))[2:]
            result += "0" + final_color if len(final_color) == 1 else final_color
        else:
            final_color = int(low + value * difference)
            result.append(final_color)
    return result



from .environments_index import EnvironmentIndex
from .maps.maps_index import MapsIndex
from .grid_world import *
from .point_env import *

# Import ant maze ONLY if mujoco_py is installed
try:
    from .ant_maze import AntMaze
except ImportError:
    pass

from .environment import Environment
from .goal_reaching_environment import GoalReachingEnv
from .navigation_environments  import *



import pickle
from copy import deepcopy
import numpy as np
import torch
from torch.nn import ReLU, Tanh
from hbrl.agents.continuous.ddpg import DDPG
from hbrl.agents.utils.mlp import MLP
from torch import optim
import torch.nn.functional as F
from gym.spaces import Box, Discrete
from typing import Union

from hbrl.utils import create_dir


class DistributionalDDPG(DDPG):
    """
    A distributional RL version of DDPG as defined in SORB (Eysenbach, 2017).
    """

    name = "Distributional DDPG"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)

        # Set up parameters as described in sorb paper
        self.batch_size = params.get("batch_size", 64)
        self.replay_buffer_max_size = params.get("replay_buffer_max_size", 1000000)
        self.actor_lr = params.get("actor_lr", 1e-4)
        self.critic_lr = params.get("critic_lr", 1e-4)
        self.tau = params.get("tau", 0.05)
        self.noise_std = params.get("noise_std", 0.1)
        self.steps_before_target_update = params.get("steps_before_target_update", 5)
        self.nb_critics = params.get("nb_critics", 1)
        self.out_dist_size = params.get("output_distribution_size", 20)
        self.out_dist_abscissa = - np.linspace(0, self.out_dist_size, self.out_dist_size)
        self.out_dist_abscissa_steps = self.out_dist_size / (self.out_dist_size - 1)

        actor_layers = params.get("actor_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        actor_activation = params.get("actor_activation", Tanh())
        assert isinstance(actor_activation, torch.nn.Module)
        self.actor = MLP(self.state_size, *actor_layers, self.nb_actions, actor_activation,
                         learning_rate=self.actor_lr, optimizer_class=optim.Adam, device=self.device).float()
        self.target_actor = deepcopy(self.actor)

        critic_layers = params.get("critic_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        self.critics = [
            MLP(self.state_size + self.nb_actions, *critic_layers, self.out_dist_size, learning_rate=self.critic_lr,
                optimizer_class=optim.Adam, device=self.device).float() for _ in range(self.nb_critics)
        ]
        self.target_critics = [deepcopy(self.critics[i]) for i in range(self.nb_critics)]

    def get_value(self, observations, actions=None, use_target=False, gradient=False):
        with torch.set_grad_enabled(gradient):
            if actions is None:
                actions = self.actor(observations).detach().numpy()
            critics = self.target_critics if use_target else self.critics
            critic_inputs = np.concatenate((observations, actions), -1)
            q_values_probs = [F.softmax(critic(critic_inputs), dim=-1) for critic in critics]
            q_values_probs = torch.stack(q_values_probs).min(0).values
            q_values = (q_values_probs * torch.tensor(self.out_dist_abscissa)).sum(dim=-1)
        return q_values if gradient else q_values.detach().cpu().numpy()

    def learn(self):
        assert not self.under_test
        if len(self.replay_buffer) > self.batch_size:
            states, actions, rewards, new_states, dones = self.sample_training_batch()

            with torch.no_grad():
                target_actions = self.target_actor.forward(new_states)
                target_critic_inputs = np.concatenate((new_states, target_actions), -1)
                target_q_values_probabilities = [F.softmax(critic(target_critic_inputs), dim=-1)
                                                 for critic in self.target_critics]
                target_q_values_probabilities = torch.stack(target_q_values_probabilities).mean(0)

            # Compute target

            # Compute distribution target
            # - Target distributions for samples where the goal has been reached
            #   shape: [[1, 0, 0, ...],
            #           [1, 0, 0, ...], ... ]
            reached_target_distribution = F.one_hot(torch.zeros(self.batch_size).to(dtype=torch.long),
                                                    self.out_dist_size)

            # - Target distribution for samples where the goal hasn't been reached
            #    * Build a column of 0
            first_column = torch.zeros(self.batch_size).unsqueeze(1)
            #    * Build others columns by shifting the target q_values
            middle_columns = target_q_values_probabilities[:, :-2]
            #    * Build the last column as the sum of the last two left q_values
            last_column = target_q_values_probabilities[:, -2:].sum(-1).unsqueeze(1)
            failed_target_distribution = torch.concat((first_column, middle_columns, last_column), -1)

            # Compute the target distributions
            target_distribution = torch.where(dones[:, np.newaxis] == 1.,
                                              reached_target_distribution,
                                              failed_target_distribution)

            for critic_id in range(self.nb_critics):
                critic = self.critics[critic_id]
                critic_distribution = critic(torch.concat((states, actions), dim=-1))
                critic_loss = F.cross_entropy(critic_distribution, target_distribution)
                critic.learn(critic_loss)
                self.target_critics[critic_id].converge_to(critic, tau=self.tau)

            # Train actor
            actions = self.actor(states)
            actor_input = torch.concat((states, actions), dim=-1)
            q_values_probs = [F.softmax(critic(actor_input), dim=-1) for critic in self.critics]
            q_values_probs = torch.stack(q_values_probs).min(0).values
            q_values = (q_values_probs * torch.tensor(self.out_dist_abscissa)).sum(dim=-1)
            self.actor.learn(torch.mean(- q_values))
            self.target_actor.converge_to(self.actor, tau=self.tau)

    def save(self, directory):
        if directory[-1] != "/":
            directory += "/"
        directory = directory[:-1]
        directory += "/"
        create_dir(directory)

        for critic_id, critic in enumerate(self.critics):
            torch.save(critic, directory + "critic_" + str(critic_id) + ".pt")

        for target_critic_id, target_critic in enumerate(self.target_critics):
            torch.save(target_critic, directory + "target_critic_" + str(target_critic_id) + ".pt")

        torch.save(self.actor, directory + "actor.pt")
        torch.save(self.target_actor, directory + "target_actor.pt")

        with open(directory + "replay_buffer.pkl", "wb") as f:
            pickle.dump(self.replay_buffer, f)

    def load(self, directory):
        if directory[-1] != "/":
            directory += "/"
        directory = directory[:-1]
        directory += "_" + str(self.nb_models) + "/"
        create_dir(directory)

        for critic_id, critic in enumerate(self.critics):
            self.critics[critic_id] = torch.load(directory + "critic_" + str(critic_id) + ".pt")

        for target_critic_id, target_critic in enumerate(self.target_critics):
            self.target_critics[target_critic_id] = \
                torch.load(directory + "target_critic_" + str(target_critic_id) + ".pt")

        self.actor = torch.load(directory + "actor.pt")
        self.target_actor = torch.load(directory + "target_actor.pt")

        with open(directory + "replay_buffer.pkl", "rb") as f:
            self.replay_buffer = pickle.load(f)



# Goal conditioned agent
from typing import Union

import numpy as np
from gym.spaces import Box, Discrete
from hbrl.agents.agent import Agent


class GoalConditionedAgent(Agent):
    """
    A global agent class for goal conditioned agents. The # NEW tag indicate differences between Agent class and this
    one.
    """

    name = "Default goal conditioned agent"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        Agent.__init__(self, state_space, action_space, **params)
        self.current_goal = None

        # Compute out goal space
        self.goal_space = params.get("goal_space", self.state_space)
        assert isinstance(self.goal_space, Box) or isinstance(self.goal_space, Discrete)
        self.goal_size = self.goal_space.shape[0]
        self.goal_shape = self.goal_space.shape
        assert len(self.goal_shape) == 1, "Multi dimensional spaces are not supported."

        # Compute state_to_goal_filter, (state[self.state_to_goal_filter] = <goal associated to state>) aka a projection
        default_filter = np.array([1] * self.goal_size + [0] * (self.state_size - self.goal_size)).astype(bool)
        self.state_to_goal_filter = params.get("state_to_goal_filter", default_filter)
        assert np.argwhere(self.state_to_goal_filter).shape[0] == self.goal_size

    def start_episode(self, state: np.ndarray, goal: np.ndarray, test_episode=False):
        super().start_episode(state, test_episode)
        self.current_goal = goal

    def reset(self):
        self.__init__(self.state_space, self.action_space, **self.init_params)



import numpy as np
import torch
from torch import nn, optim
from torch.nn.modules import Module
from torch.optim import Optimizer

class NN(Module):
    def __init__(self):
        super().__init__()

    def converge_to(self, other_mlp, tau=0.1):
        """
        Make the weights of the current model be a bit closer to the given mlp.
        self.weights = (1 - tau) * self.weights + tau * other_mlp.weights
        Precondition: other_mlp have the exact same shape of self.
        """
        for self_param, other_param in zip(self.parameters(), other_mlp.parameters()):
            self_param.data.copy_(
                self_param.data * (1.0 - tau) + other_param.data * tau
            )

    def learn(self, loss, retain_graph=False):
        self.optimizer.zero_grad()
        loss.backward(retain_graph=retain_graph)
        self.optimizer.step()



class MLP(NN):
    """
    A general MLP class. Initialisation example:
    mlp = MLP(input_size, 64, ReLU(), 64, ReLU(), output_size, Sigmoid())
    """

    def __init__(self, input_size, *layers_data, device, learning_rate=0.01,
                 optimizer_class=optim.Adam):
        """
        For each element in layers_data:
         - If the element is an integer, it will be replaced by a linear layer with this integer as output size,
         - If this is a model (like activation layer) il will be directly integrated
         - If it is a function, it will be used to initialise the weights of the layer before
            So we call layer_data[n](layer_data[n - 1].weights) with n the index of the activation function in
            layers_data
        """
        super().__init__()
        assert issubclass(optimizer_class, Optimizer)
        assert device is not None

        self.layers = nn.ModuleList()
        self.input_size = input_size  # Can be useful later ...
        for data in layers_data:
            layer = data
            if isinstance(data, int):
                layer = nn.Linear(input_size, data)
                input_size = data
            if callable(data) and not isinstance(data, nn.Module):
                data(self.layers[-1].weight)
                continue
            self.layers.append(layer)  # For the next layer

        self.device = device
        self.to(self.device)
        self.learning_rate = learning_rate
        self.optimizer = optimizer_class(params=self.parameters(), lr=learning_rate)

    def forward(self, input_data):
        if isinstance(input_data, np.ndarray):
            input_data = torch.from_numpy(input_data).to(self.device)
        input_data = input_data.float()
        for layer in self.layers:
            input_data = layer(input_data)
        return input_data



import math
import random
from .mujoco_files.maps.maps_index import TileType
import numpy as np
from gym.spaces import Box
from mujoco_py import load_model_from_path, MjSim, MjViewer
from scipy.spatial import distance
from .mujoco_files.xml_generator import generate_xml
from .mujoco_model_utils import quat2mat, euler2quat, mat2euler

"""
SETTINGS
"""

MAZE_TILE_RESOLUTION = 50  # size os a tile of the maze_grid in pixels


class AntMaze:

    def __init__(self, angle=0, maze_name="empty_room", image_resolution_per_tile=50, show=False):
        """
        Initialise an ant maze environment.
        THe model is automatically created using a map specification defined in /mujoco_files/maps/
        The maze_name should correspond to a map name, so /mujoco_files/maps/<maze_name>.txt should exist.
        """
        self.angle = angle
        self.maze_name = maze_name
        self.image_resolution_per_tile = image_resolution_per_tile
        self.maze_array, xml_spec_path = generate_xml(maze_name)  # TODO : check maze_info["reachable_spaces_size"]
        self.maze_array = np.array(self.maze_array)
        self.maze_array_height, self.maze_array_width = self.maze_array.shape

        # Create Mujoco Simulation
        self.model = load_model_from_path(xml_spec_path)
        self.sim = MjSim(self.model)

        low = self.map_coordinates_to_env_position(0, -1)
        high = self.map_coordinates_to_env_position(-1, 0)
        self.maze_space = Box(low=low, high=high)  # Verify low and high
        self.goal_space = Box(low=np.append(low, 0.45), high=np.append(high, 0.55))
        self.goal_size = self.goal_space.shape[0]

        # Observation space
        self.state_size = len(self.sim.data.qpos) + len(self.sim.data.qvel)
        fill_size = self.state_size - self.maze_space.shape[0]
        observation_low = np.concatenate((self.maze_space.low, np.full(fill_size, float("-inf"))))
        observation_high = np.concatenate((self.maze_space.high, np.full(fill_size, float("inf"))))
        self.state_space = Box(low=observation_low, high=observation_high)

        # Action space
        self.action_space = Box(low=self.sim.model.actuator_ctrlrange[:, 0],
                                high=self.sim.model.actuator_ctrlrange[:, 1])
        self.action_size = self.action_space.shape[0]

        max_actions = 500
        num_frames_skip = 10

        self.goal_thresholds = np.array([0.5, 0.5, 0.2, 0.5, 0.5])

        self.max_actions = max_actions

        # Implement visualization if necessary
        self.show = show
        if self.show:
            self.viewer = MjViewer(self.sim)
        self.num_frames_skip = num_frames_skip

        self.goal = None

    # Get state, which concatenates joint positions and velocities
    def get_state(self):
        return np.concatenate((self.sim.data.qpos, self.sim.data.qvel))

    def map_coordinates_to_env_position(self, x, y):
        x = self.maze_array_width + x if x < 0 else x
        y = self.maze_array_height + y if y < 0 else y
        return np.array([x - self.maze_array_width / 2 + 0.5, - (y - self.maze_array_height / 2 + 0.5)])

    def sample_reachable_position(self):
        # sample empty tile
        tile_coordinates = np.flip(random.choice(np.argwhere(self.maze_array != TileType.WALL.value)))
        position = self.map_coordinates_to_env_position(*tile_coordinates)

        # Sample a random position in the chosen tile
        low, high = position - 0.5, position + 0.5
        return np.random.uniform(low, high)

    # Reset simulation to state within initial state specified by user
    def reset(self):

        # Reset controls
        self.sim.data.ctrl[:] = 0

        # Reset joint positions and velocities

        # Sample a start position
        start_tiles = np.argwhere(self.maze_array == TileType.START.value)
        start_coordinates = np.flip(random.choice(start_tiles))
        start_position = self.map_coordinates_to_env_position(*start_coordinates)

        # Build the state from it, using fixed joints and velocities values.
        angle = random.random() * 2 - 1  # Sample random orientation for the beginning
        initial_qpos = np.array([0.55, 1.0, 0.0, 0.0, angle, 0.0, 1.0, 0.0, -1.0, 0.0, -1.0, 0.0, 1.0])
        initial_qvel = np.zeros(14)
        state = np.concatenate((start_position, initial_qpos, initial_qvel))

        self.sim.data.qpos[:] = state[:len(self.sim.data.qpos)]
        self.sim.data.qvel[:] = state[len(self.sim.data.qpos):]
        self.sim.step()

        # Choose a goal. Weights are used to keep a uniform sampling over the reachable space,
        # since boxes don't cover areas of the same size.
        goal_position = self.sample_reachable_position()
        torso_height = np.random.uniform(0.45, 0.55, (1,))
        self.goal = np.concatenate((goal_position, torso_height))
        if self.show:
            self.sim.data.mocap_pos[0][:2] = np.copy(self.goal[:2])

        if self.show:
            self.viewer.render()
        # Return state
        state = self.get_state()
        return state, self.goal

    # Execute low-level action for number of frames specified by num_frames_skip
    def step(self, action):
        self.sim.data.ctrl[:] = action
        for _ in range(self.num_frames_skip):
            self.sim.step()
            if self.show:
                self.viewer.render()
        new_state = self.get_state()

        reached = (np.abs(new_state[:len(self.goal)] - self.goal) < self.goal_thresholds[:len(self.goal)]).all()
        return self.get_state(), 0 if reached else -1, reached

    # Visualize end goal. This function may need to be adjusted for new environments.
    def display_end_goal(self):
        self.sim.data.mocap_pos[0][:2] = np.copy(self.goal[:2])

    # Visualize all sub-goals
    def display_sub_goals(self, sub_goals):

        # Display up to 10 sub-goals and end goal
        if len(sub_goals) <= 11:
            sub_goal_ind = 0
        else:
            sub_goal_ind = len(sub_goals) - 11

        for i in range(1, min(len(sub_goals), 11)):
            self.sim.data.mocap_pos[i][:2] = np.copy(sub_goals[sub_goal_ind][:2])
            self.sim.model.site_rgba[i][3] = 1

            sub_goal_ind += 1

    def set_node(self, node_name, rgba=None, position=None):
        geom_id = self.sim.model.geom_name2id(node_name)

        if rgba is not None:
            if isinstance(rgba, str):
                rgba = np.array([float(elt) for elt in rgba.split(" ")])
            if isinstance(rgba, list):
                rgba = np.array(rgba)
            self.sim.model.geom_rgba[geom_id, :] = rgba

        if position is not None:
            if isinstance(position, list):
                position = np.array(position)
            assert len(position) <= 3
            self.sim.model.geom_pos[geom_id, :len(position)] = position

    def set_edge(self, edge_name, first_node: np.ndarray, second_node: np.ndarray, rgba=None):
        edge_id = self.sim.model.geom_name2id(edge_name)

        assert first_node.shape == (2,)
        assert second_node.shape == (2,)

        # Compute edge position
        position = (first_node + second_node) / 2
        self.sim.model.geom_pos[edge_id] = position

        # Compute angle between these two nodes
        diff = second_node - first_node
        angle = math.acos(diff[0] / np.linalg.norm(diff))
        euler_rotation = np.array([-math.pi / 2, angle, 0])
        self.sim.model.geom_quat[edge_id] = euler2quat(euler_rotation)

        if rgba is not None:
            assert isinstance(rgba, str) and len(rgba.split(" ")) <= 4
            self.sim.model.geom_rgba[edge_id] = rgba

    def toggle_geom_object(self, object_name):
        geom_id = self.sim.model.geom_name2id(object_name)
        self.sim.model.geom_rgba[geom_id, -1] = 1 - self.sim.model.geom_rgba[geom_id, -1]

    def set_space_color(self, image_array: np.ndarray, low, high, color) -> np.ndarray:
        """
        Set a tile color with the given color in the given image as a numpy array of pixels
        :param image_array: The image where the tile should be set
        :param low: the box lower corner
        :param high: the box higher corner
        :param color: new color of the tile : numpy array [Red, Green, Blue]
        :return: The new image
        """
        tile_size = MAZE_TILE_RESOLUTION
        tile_width, tile_height = np.rint((high - low) * tile_size).astype(int)
        tile_img = np.tile(color, (tile_height, tile_width, 1))

        image_width, image_height, _ = image_array.shape

        # We add +1 to take images borders in consideration. Without it, x/y_min positions are computed from the
        # border of self.maze_space, that do not include maze borders.
        # NB: Y coordinates don't work the same in the state space and in the image.
        # High values (in the state space) are at the top of the image (low numpy index), with low coordinates in
        # the image. Opposite for X values.
        x_min = round((low[0] - self.maze_space.low[0] + 1) * tile_size)
        y_max = round(((image_height / tile_size) - (low[1] - self.maze_space.low[1] + 1)) * tile_size)

        x_max = x_min + tile_width
        y_min = y_max - tile_height

        image_array[y_min:y_max, x_min:x_max, :] = tile_img
        return image_array

    def render(self) -> np.ndarray:
        """
        Return a np.ndarray of size (width, height, 3) of pixels that represent the environments and it's walls
        :return: The final image.
        """

        # we add np.full(2, 2) to get image borders, because border walls are not included in env.maze_space.
        # We assume walls have a width of one, but it's not really important for a generated image.
        image_width_px, image_height_px = np.rint((self.maze_space.high - self.maze_space.low + np.full(2, 2))
                                                  * self.image_resolution_per_tile).astype(int)

        img = np.zeros(shape=(image_height_px, image_width_px, 3), dtype=np.uint8)

        # Render the grid
        for coordinates in np.argwhere(self.maze_array != TileType.WALL.value):
            coordinates = np.flip(coordinates)
            position = self.map_coordinates_to_env_position(*coordinates)
            img = self.set_space_color(img, position - 0.5, position + 0.5, np.array([255, 255, 255]))

        self.place_point(img, self.sim.data.qpos[:2], [0, 0, 255], width=10)
        self.place_point(img, self.goal[:2], [255, 0, 0], width=10)
        return img

    def place_point(self, image: np.ndarray, state, color: np.ndarray, width=50):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        x, y = tuple(state[:2])
        tile_size = MAZE_TILE_RESOLUTION

        # Convert x, y state_space_position into maze_coordinates to get point center
        # Explanation: same as in self.set_space_color, lines of code 5/6
        x_center_px = round((x - self.maze_space.low[0] + 1) * tile_size)
        y_center_px = round(((image.shape[1] / tile_size) - (y - self.maze_space.low[1] + 1)) * tile_size)

        # Imagine a square of size width * width, with the coordinates computed above as a center. Iterate through
        # each pixel inside this square to
        radius = round(width / 2) + 1
        for i in range(x_center_px - radius, x_center_px + radius):
            for j in range(y_center_px - radius, y_center_px + radius):
                if distance.euclidean((i + 0.5, j + 0.5), (x_center_px, y_center_px)) < radius:
                    image[j, i] = color

    def place_edge(self, image: np.ndarray, state_1, state_2, color: np.ndarray, width=40):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        tile_size = MAZE_TILE_RESOLUTION
        x1, y1 = tuple(state_1[:2])
        x2, y2 = tuple(state_2[:2])

        # Convert x, y state_space_position into maze_coordinates to get point center
        # Explanation: same as in self.set_space_color, lines of code 5/6
        x1_center_px = (x1 - self.maze_space.low[0] + 1) * tile_size
        y1_center_px = ((image.shape[1] / tile_size) - (y1 - self.maze_space.low[1] + 1)) * tile_size
        x2_center_px = (x2 - self.maze_space.low[0] + 1) * tile_size
        y2_center_px = ((image.shape[1] / tile_size) - (y2 - self.maze_space.low[1] + 1)) * tile_size
        x_min = int(min(x1_center_px, x2_center_px))
        x_max = int(max(x1_center_px, x2_center_px))
        y_min = int(min(y1_center_px, y2_center_px))
        y_max = int(max(y1_center_px, y2_center_px))
        # Imagine a square of size width * width, with the coordinates computed above as a center. Iterate through
        # each pixel inside this square to
        for i in range(x_min, x_max):
            for j in range(y_min, y_max):
                cross_product = (y2_center_px - y1_center_px) * (i - x1_center_px) \
                    - (x2_center_px - x1_center_px) * (j - y1_center_px)
                # compare versus epsilon for floating point values, or != 0 if using integers
                if abs(cross_product) > width * 10:
                    continue
                image[j, i] = color



# Goal conditioned deep Q-network

import copy
import pickle

import numpy as np
import torch
from gym.spaces import Box
from torch.nn import ReLU, Tanh
from torch.nn.functional import normalize

from hbrl.agents.utils.mlp import MLP
from hbrl.agents.value_based_agent import ValueBasedAgent
from torch import optim
from torch.nn import functional
from torch.distributions.normal import Normal
from gym.spaces import Box, Discrete
from typing import Union

from hbrl.utils import create_dir


class SAC(ValueBasedAgent):

    name = "SAC"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)

        self.actor_lr = params.get("actor_lr", 0.0005)
        self.critic_lr = params.get("critic_lr", 0.0005)
        alpha = params.get("alpha", None)
        self.critic_alpha = params.get("critic_alpha", 0.6)
        self.actor_alpha = params.get("actor_alpha", 0.05)
        if alpha is not None:
            self.critic_alpha = alpha
            self.actor_alpha = alpha
        self.gamma = params.get("gamma", 0.99)
        self.tau = params.get("tau", 0.005)
        self.reward_scale = params.get("reward_scale", 15)

        self.policy_update_frequency = 2
        self.learning_step = 1

        self.min_std = -20
        self.max_std = 2

        actor_layers = params.get("actor_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        actor_activation = params.get("actor_activation", Tanh())
        assert isinstance(actor_activation, torch.nn.Module)
        self.actor = MLP(self.state_size, *actor_layers, 2 * self.nb_actions, actor_activation,
                         learning_rate=self.actor_lr, optimizer_class=optim.Adam, device=self.device).float()
        self.target_actor = copy.deepcopy(self.actor)

        critic_layers = params.get("critic_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        self.critic = MLP(self.state_size + self.nb_actions, *critic_layers, 1,
                          learning_rate=self.critic_lr, optimizer_class=optim.Adam, device=self.device).float()
        self.target_critic = copy.deepcopy(self.critic)

        self.passed_logs = []

    def get_q_value(self, state):
        with torch.no_grad():
            state = torch.from_numpy(state).to(self.device) if isinstance(state, np.ndarray) else state

            next_actions, _ = self.sample_action(state, use_target_network=True)
            critic_input = torch.concat((state, next_actions), dim=-1)
            q_values = self.target_critic.forward(critic_input).view(-1)
        return q_values

    def sample_action(self, actor_input, use_target_network=False, explore=True):
        actor_network = self.target_actor if use_target_network else self.actor

        if isinstance(actor_input, np.ndarray):
            actor_input = torch.from_numpy(actor_input).to(self.device)
        actor_input = normalize(actor_input, p=2., dim=-1) # Tensor torch.float64

        # Forward
        actor_output = actor_network(actor_input)
        if len(actor_input.shape) > 1:  # It's a batch
            actions_means = actor_output[:, :self.nb_actions]
            actions_log_stds = actor_output[:, self.nb_actions:]
        else:
            actions_means = actor_output[:self.nb_actions]
            actions_log_stds = actor_output[self.nb_actions:]

        actions_log_stds = torch.clamp(actions_log_stds, min=self.min_std, max=self.max_std)
        actions_stds = torch.exp(actions_log_stds)
        actions_distribution = Normal(actions_means, actions_stds)

        raw_actions = actions_distribution.rsample() if explore or self.under_test else actions_means

        log_probs = actions_distribution.log_prob(raw_actions)
        actions = torch.tanh(raw_actions)
        log_probs -= torch.log(1 - actions.pow(2) + 1e-6)
        log_probs = log_probs.sum(-1)
        actions = self.scale_action(actions, Box(-1, 1, (self.nb_actions,)))
        return actions, log_probs

    def action(self, state, explore=True):
        with torch.no_grad():
            action, _ = self.sample_action(state, explore=explore)
        return action.cpu().detach().numpy()

    def get_value(self, observations, actions=None):
        with torch.no_grad():
            if actions is None:
                actions, _ = self.sample_action(observations)
            elif isinstance(actions, np.ndarray):
                actions = torch.tensor(actions)
            if isinstance(observations, np.ndarray):
                observations = torch.tensor(observations)
            critic_value = self.critic(torch.concat((observations, actions), dim=-1))
        if len(critic_value.shape) > 1:
            critic_value = critic_value.squeeze()
        return critic_value.detach().numpy()

    def learn(self):
        if not self.under_test and len(self.replay_buffer) > self.batch_size:
            states, actions, rewards, new_states, done = self.sample_training_batch()

            # Training critic
            with torch.no_grad():
                next_actions, next_log_probs = \
                    self.sample_action(new_states, use_target_network=True)
                critic_input = torch.concat((new_states, next_actions), dim=-1)
                self.passed_logs.append(next_log_probs)
                next_q_values = \
                    self.target_critic(critic_input).view(-1)

            q_hat = self.reward_scale * rewards + self.gamma * (1 - done) * \
                (next_q_values - self.critic_alpha * next_log_probs)
            q_values = self.critic(torch.concat((states, actions), dim=-1)).view(-1)
            critic_loss = functional.mse_loss(q_values, q_hat)
            self.critic.learn(critic_loss)
            self.target_critic.converge_to(self.critic, tau=self.tau)

            if self.learning_step % self.policy_update_frequency == 0:
                for _ in range(self.policy_update_frequency):
                    # Train actor
                    actions, log_probs = self.sample_action(states)
                    log_probs = log_probs.view(-1)
                    critic_values = self.critic(torch.concat((states, actions), dim=-1)).view(-1)

                    actor_loss = self.actor_alpha * log_probs - critic_values
                    actor_loss = torch.mean(actor_loss)
                    self.actor.learn(actor_loss, retain_graph=True)
                    self.target_actor.converge_to(self.actor, tau=self.tau)
            self.learning_step += 1

    def save(self, directory):
        if directory[-1] != "/":
            directory += "/"
        create_dir(directory)

        torch.save(self.critic, directory + "critic.pt")
        torch.save(self.target_critic, directory + "target_critic.pt")

        torch.save(self.actor, directory + "actor.pt")
        torch.save(self.target_actor, directory + "target_actor.pt")

        with open(directory + "replay_buffer.pkl", "wb") as f:
            pickle.dump(self.replay_buffer, f)

    def load(self, directory):
        if directory[-1] != "/":
            directory += "/"
        directory = directory[:-1]
        directory += "/"
        create_dir(directory)

        self.critic = torch.load(directory + "critic.pt")
        self.target_critic = torch.load(directory + "target_critic.pt")

        self.actor = torch.load(directory + "actor.pt")
        self.target_actor = torch.load(directory + "target_actor.pt")

        with open(directory + "replay_buffer.pkl", "rb") as f:
            self.replay_buffer = pickle.load(f)


from enum import Enum


class Direction(Enum):
    TOP = 0
    BOTTOM = 1
    LEFT = 2
    RIGHT = 3


class TileType(Enum):
    EMPTY = 0
    WALL = 1
    START = 2
    TERMINAL = 3


class Colors(Enum):
    EMPTY = [250, 250, 250]
    WALL = [50, 54, 51]
    START = [213, 219, 214]
    TERMINAL = [73, 179, 101]
    TILE_BORDER = [50, 54, 51]
    AGENT = [0, 0, 255]
    GOAL = [255, 0, 0]



# Goal conditioned agent
from typing import Union

import numpy as np
from gym.spaces import Box, Discrete

from hbrl.agents.goal_conditioned_wrappers.goal_conditioned_agent import GoalConditionedAgent
from hbrl.agents.value_based_agent import ValueBasedAgent


class GoalConditionedValueBasedAgent(GoalConditionedAgent, ValueBasedAgent):
    """
    A global agent class for goal conditioned agents. The # NEW tag indicate differences between Agent class and this
    one.
    """

    def __init__(self, reinforcement_learning_agent_class, state_space: Union[Box, Discrete],
                 action_space: Union[Box, Discrete], **params):
        GoalConditionedAgent.__init__(self, state_space, action_space, **params)
        self.reinforcement_learning_agent_class = reinforcement_learning_agent_class

        # Compute our agent new state space as a goal-conditioned state space (a concatenation of our state space and
        # our goal space)
        self.reinforcement_learning_agent: ValueBasedAgent = \
            reinforcement_learning_agent_class(self.feature_space, action_space, **params)

    def __getattr__(self, name):
        """Returns an attribute with ``name``, unless ``name`` starts with an underscore."""
        if name.startswith("_"):
            raise AttributeError(f"accessing private attribute '{name}' is prohibited")
        return getattr(self.reinforcement_learning_agent, name)

    def learn(self):
        self.reinforcement_learning_agent.learn()

    @property
    def feature_space(self):
        if isinstance(self.state_space, Box) and isinstance(self.goal_space, Box):
            return Box(low=np.concatenate((self.state_space.low, self.goal_space.low)),
                       high=np.concatenate((self.state_space.high, self.goal_space.high)))
        elif isinstance(self.state_space, Discrete) and isinstance(self.goal_space, Discrete):
            return Discrete(self.state_space.n * self.goal_space.n)
        else:
            raise NotImplementedError("State space ang goal space with different types are not supported.")

    def get_features(self, states, goals):
        if isinstance(self.state_space, Box) and isinstance(self.goal_space, Box):
            return np.concatenate((states, goals), -1)
        elif isinstance(self.state_space, Discrete) and isinstance(self.goal_space, Discrete):
            return states + goals * self.state_space.n  # Use a bijection between N² and N
        else:
            raise NotImplementedError("State space ang goal space with different types are not supported.")

    def get_value(self, state, goal, actions=None):
        return self.reinforcement_learning_agent.get_value(self.get_features(state, goal), actions)

    def start_episode(self, state: np.ndarray, goal: np.ndarray, test_episode=False):
        GoalConditionedAgent.start_episode(self, state, goal, test_episode)
        self.reinforcement_learning_agent.start_episode(self.get_features(state, self.current_goal), test_episode)

    def action(self, state, explore=True):
        return self.reinforcement_learning_agent.action(self.get_features(state, self.current_goal), explore)

    def process_interaction(self, action, reward, new_state, done, learn=True):
        super().process_interaction(action, reward, new_state, done, learn=learn)
        new_state = self.get_features(new_state, self.current_goal)
        self.reinforcement_learning_agent.process_interaction(action, reward, new_state, done, learn)

    def save_interaction(self, state, action, reward, new_state, done, goal=None):
        assert not self.under_test
        goal = self.current_goal if goal is None else goal
        state = self.get_features(state, goal)
        new_state = self.get_features(new_state, goal)
        self.reinforcement_learning_agent.save_interaction(state, action, reward, new_state, done)

    def get_estimated_distances(self, states, goals):
        """
        Return the estimated distance between given goals and states.
        """
        features = self.get_features(states, goals)
        estimated_distance = self.reinforcement_learning_agent.get_value(features)
        if len(estimated_distance.shape) == 0:
            estimated_distance = estimated_distance.reshape((1,))
        estimated_distance = - estimated_distance.clip(float("-inf"), 0)
        return estimated_distance

    def reset(self):
        self.__init__(self.reinforcement_learning_agent_class, self.state_space, self.action_space, **self.init_params)

if __name__ == "__main__":
    from hbrl.agents.discrete.dqn import DQN
    from hbrl.environments.grid_world.goal_conditioned_discrete_grid_world import GoalConditionedDiscreteGridWorld

    environment = GoalConditionedDiscreteGridWorld()
    dqn_agent = DQN(action_space=environment.action_space, state_space=environment.state_space)

    new_agent = GoalConditionedValueBasedAgent(DQN, action_space=environment.action_space, state_space=environment.state_space)
    same_buffers = id(new_agent.replay_buffer) == id(new_agent.reinforcement_learning_agent.replay_buffer)

    state, goal = environment.reset()
    print("state: ", state, "; goal: ", goal, "; features: ", new_agent.get_features(state, goal))

    new_agent.start_episode(state, goal)
    print(new_agent.action(state))

    a = 1





def get_dict_as_str(input_dict, start_tab="", tab="    "):
    result = ""
    for k, v in input_dict.items():
        if result != "":
            result += ",\n"
        if isinstance(v, dict):
            result += get_dict_as_str(v, start_tab + tab)
        if isinstance(v, list):
            result += start_tab + tab + str(k) + ": [" + "\n"
            for index, elt in enumerate(v):
                if isinstance(elt, dict):
                    result += get_dict_as_str(elt, start_tab + tab + tab)
                else:
                    result += start_tab + tab + tab + str(elt)
                result += "\n" if index == len(v) - 1 else ",\n"
            result += start_tab + tab + "]"
        else:
            if isinstance(v, str):
                result += start_tab + tab + str(k) + ": \"" + v + "\""
            else:
                result += start_tab + tab + str(k) + ": " + str(v) + ""
    return start_tab + "{\n" + result + "\n" + start_tab + "}"


from typing import Union

import numpy as np
import torch
from gym.spaces import Box


def signlog(target):
    """
    This function can be applied on the critic target to help convergence.
    It scales down the critic target value, so it's no more necessary to normalise the reward.
    This function is similar to a Tanh, but it is not bounded between -1 and 1, so when the target q-value is high,
    signlog(target) can still be different enough from signlog(target - 1).

    usage:
    self.critic(state, action,

    """
    return torch.sign(target) * torch.log(target + 1)


def scale_tensor(data: Union[np.ndarray, torch.Tensor], input_space: Box, output_space: Box):
    """
    Scale an action within the given bounds action_low to action_high, to our action_space.
    The result action is also clipped to fit in the action space in case the given action wasn't exactly inside
    the given bounds.
    Useless if our action space is discrete.
    @return: scaled and clipped actions. WARNING: actions are both attribute and result. They are modified by the
    function. They are also returned for better convenience.
    """
    assert isinstance(data, (np.ndarray, torch.Tensor))
    assert isinstance(input_space, Box)
    assert isinstance(output_space, Box)

    source_low, source_high = input_space.low, input_space.high
    target_low, target_high = output_space.low, output_space.high
    if isinstance(data, torch.Tensor):
        source_low, source_high = torch.tensor(input_space.low), torch.tensor(source_high)
        target_low, target_high = torch.tensor(target_low), torch.tensor(target_high)

    # Scale action to the action space
    source_range = source_high - input_space.low
    target_range = target_high - target_low

    scale = target_range / source_range
    data = data * scale
    data = data + (target_low - (input_space.low * scale))
    clip_fun = np.clip if isinstance(data, np.ndarray) else torch.clamp
    data = clip_fun(data, target_low, target_high)
    return data



import abc
import copy
from typing import Union
from warnings import warn

import numpy as np
from gym.spaces import Box, Discrete
from abc import ABC


class Environment(ABC):

    name = "Default Environment"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete]):
        assert len(state_space.shape) == 1, "Environments with state space with multiple dimensions are not supported. " \
                                            "Flatten your state space or use another library."
        assert isinstance(state_space, Box) or isinstance(state_space, Discrete), \
            "State spaces with a different type than gym.spaces.Box or gym.spaces.Discrete ar not supported."
        self.state_space = state_space
        self.state_size = state_space.shape[0] if isinstance(state_space, Box) else state_space.n
        self.state_space_bounds = self.state_space.high
        self.state_space_offset = (self.state_space.high + self.state_space.low) / 2

        assert len(action_space.shape) == 1, "Environments with action space with multiple dimensions are not supported. " \
                                            "Flatten your action space or use another library."
        assert isinstance(action_space, Box) or isinstance(action_space, Discrete), \
            "Action spaces with a different type than gym.spaces.Box or gym.spaces.Discrete ar not supported."
        self.action_space = action_space
        self.nb_actions = action_space.shape[0] if isinstance(action_space, Box) else action_space.n
        self.action_space_bounds = self.action_space.high
        self.action_space_offset = (self.action_space.high + self.action_space.low) / 2

        self.episode_duration = 0  # Nb steps taken in the current episode
        self.state = None
        self.reset()

    def reset(self) -> np.ndarray:
        self.state = self.state_space.sample()
        self.episode_duration = 0
        return self.state

    def step(self, action: np.ndarray) -> (np.ndarray, float, bool):
        assert self.action_space.contains(action.astype(self.action_space.dtype))
        if not self.action_space.contains(action):
            warn("Environment received an action with stata-type" + str(action.dtype) +
                 " where its action space data-type is " + str(self.action_space.dtype) + ".")

        self.episode_duration += 1

        return np.random.rand(), 0.0, False

    @abc.abstractmethod
    def render(self) -> np.ndarray:
        """
        Return a numpy array that represent an image (with shape (_, _, 3)) of the environment in its current state.
        """
        pass

    def copy(self):
        """
        Returns a copy of this environment.
        """
        return copy.deepcopy(self)



import random
import numpy as np
from gym.spaces import Box, Discrete
from ...goal_conditioned_wrappers.goal_conditioned_agent import GoalConditionedAgent
from ...goal_conditioned_wrappers.goal_conditioned_value_based_agent import GoalConditionedValueBasedAgent

"""
This implementation of Hierarchical Actor-Critic (HAC) use hindsight experience replay to learn. It is an implementation
of the paper "Learning multi-level hierarchies with hindsight" from Andrew Levy et al. (ICLR 2019).
"""

class HRL(GoalConditionedAgent):
    def __init__(self, control_algorithm, nb_layers=3, high_level_agent=None, **params):
        """
        @param control_algorithm: The RL algorithm used to select the lowest level action (aka. control action).
        @param nb_layers: how many layers do we use. Should be > 1.
        @param high_level_agent: The RL algorithm used to generate sub_goals. If None, then we copy the
            control policy to get the sub-goals generator algorithm.

        NB: The sub-goal generation class can be different from the control policy because if the control action space
        is different from the raw state spaces (the ones of the lowest level) then the action space of the higher layer
        can be continuous while the action space of the lowest is discrete (for example).

        @param params: Optional parameters (check __init__ function for more details).
        """
        # Run some tests about parameters and set action_space and state_space using given agents.
        assert isinstance(control_algorithm, GoalConditionedValueBasedAgent)

        state_space = control_algorithm.state_space
        action_space = control_algorithm.action_space
        assert state_space == high_level_agent.state_space

        # Initialise super class and then initialise us.
        super().__init__(state_space, action_space, **params)

        self.name = "HRL(" + control_algorithm.name + ", " + str(nb_layers) \
                    + ("" if high_level_agent is None else ", " + high_level_agent.name) + ")"

        # Compute sub_goal space
        self.sub_goal_space = params.get("goal_space", self.state_space)
        self.sub_goal_size = self.sub_goal_space.shape[0]
        self.sub_goal_shape = self.sub_goal_space.shape
        assert isinstance(self.sub_goal_space, Box) or isinstance(self.sub_goal_space, Discrete)
        assert len(self.sub_goal_shape) == 1, "Multi dimensional spaces are not supported."

        # Store arguments in memory (params are stored by the mother class)
        self.control_algorithm = control_algorithm
        self.high_level_agent = control_algorithm.copy() if high_level_agent is None else high_level_agent.copy()
        assert isinstance(high_level_agent, GoalConditionedValueBasedAgent)
        self.high_level_agent.action_space = self.sub_goal_space
        self.high_level_agent.reset()
        # NB: other arguments are stored by the mother class.

        # Compute state_to_sub_goal_filter, (state[self.state_to_sub_goal_filter] = <goal associated to state>) aka a projection
        default_filter = np.array([1] * self.sub_goal_size + [0] * (self.state_size - self.sub_goal_size)).astype(bool)
        self.state_to_sub_goal_filter = params.get("state_to_sub_goal_filter", default_filter)
        assert np.argwhere(self.state_to_sub_goal_filter).shape[0] == self.sub_goal_size

        self.planning_horizon = params.get("planning_horizon", 20)
        self.sg_test_probability = params.get("sub_goal_test_probability", 0)
        self.layer_id = nb_layers - 1

        self.reachability_threshold = params.get("reachability_threshold", .8)

        # Layers initialisations
        if self.layer_id == 1:
            self.control_policy_trials_to_reach_sg = 0
        self.testing_sub_goal = False
        self.last_trajectory = []

        if self.layer_id == 1:
            self.lower_level = self.control_algorithm
        else:
            self.lower_level = HRL(control_algorithm, nb_layers - 1, high_level_agent, **params)

        self.done = False
        self.sub_goal = None
        self.nb_sub_goals_sampled = 0

        self.use_basic_experience_replay = True
        self.use_sub_goals_penalty = False
        self.use_hindsight_actions_replay = False
        self.use_hindsight_goal_replays = True

    @property
    def sub_goals(self):
        if self.layer_id == 1:
            return [self.sub_goal, self.current_goal]
        else:
            return self.lower_level.sub_goals + [self.current_goal]

    def get_layer(self, layer_id):
        assert layer_id >= 0
        if self.layer_id == layer_id:
            return self
        if (self.layer_id - 1) == layer_id:
            return self.lower_level
        else:
            return self.lower_level.get_layer(layer_id)

    def start_episode(self, state: np.ndarray, goal: np.ndarray, test_episode=False):
        super().start_episode(state, goal, test_episode)
        self.testing_sub_goal = False
        self.done = False
        self.nb_sub_goals_sampled = 0
        self.last_trajectory = []

        self.high_level_agent.start_episode(state, goal, test_episode)
        self.reset_sub_goal(state)

    def process_interaction(self, action, reward, new_state, done, learn=True):

        reached = np.linalg.norm(new_state - self.sub_goal) < self.reachability_threshold
        self.lower_level.process_interaction(action, 0 if reached else -1, new_state, reached or done, learn)

        if reached or (self.layer_id == 1 and self.control_policy_trials_to_reach_sg >= self.planning_horizon) \
                or (self.layer_id > 1 and self.lower_level.done):
            action = self.sub_goal
            super().process_interaction(action, reward, new_state, done)
            if learn and not self.under_test:
                self.last_trajectory.append((self.last_state, action))
                self.high_level_agent.process_interaction(action, reward, new_state, done,
                                                          learn=self.use_basic_experience_replay)
                if not self.use_basic_experience_replay:
                    self.high_level_agent.learn()

                if self.use_sub_goals_penalty:
                    if not reached and self.testing_sub_goal:
                        if np.linalg.norm(self.last_state - self.sub_goal) < 4:
                            a = 1
                        self.high_level_agent.save_interaction(self.last_state, action, -self.planning_horizon,
                                                               new_state, True)
            if self.nb_sub_goals_sampled >= self.planning_horizon:
                self.done = True
            self.reset_sub_goal(new_state)

        # sub-goals verification routine
        if self.sub_goal is None:
            print("goal is None")
        assert (self.lower_level.current_goal == self.sub_goal).all()

    def action(self, state):
        if self.layer_id == 1:
            self.control_policy_trials_to_reach_sg += 1
        return self.lower_level.action(state)

    def stop_episode(self):
        super().stop_episode()

        # For each observation seen :
        if self.use_hindsight_goal_replays or self.use_hindsight_actions_replay:
            for state_index, (state, action) in enumerate(self.last_trajectory[:-1]):
                new_state_index = state_index + 1
                new_state, _ = self.last_trajectory[new_state_index]

                hindsight_goal_index = len(self.last_trajectory) - 1
                # reward = (new_state_index / goal_index) - 1  # For a more dense reward
                reward = 0 if new_state_index == hindsight_goal_index else -1
                done = hindsight_goal_index == new_state_index

                # sample four goals in future states
                if self.use_hindsight_actions_replay:
                    hindsight_action = new_state.copy()
                    self.high_level_agent.save_interaction(state, hindsight_action, reward, new_state, done,
                                                           self.current_goal)

                if self.use_hindsight_goal_replays:
                    hindsight_goal, _ = self.last_trajectory[-1]
                    hindsight_goal = hindsight_goal[self.state_to_goal_filter]
                    self.high_level_agent.save_interaction(state, action, reward, new_state, done, hindsight_goal)

        self.high_level_agent.stop_episode()
        self.lower_level.stop_episode()

    def reset_sub_goal(self, state):
        assert self.current_goal is not None
        assert (self.current_goal == self.high_level_agent.current_goal).all()

        self.sub_goal = self.high_level_agent.action(state)
        self.nb_sub_goals_sampled += 1
        self.testing_sub_goal = self.sg_test_probability > random.random()
        if self.layer_id == 1:
            self.control_policy_trials_to_reach_sg = 0

        # Reset low level agent's episode
        self.lower_level.start_episode(state, self.sub_goal, test_episode=self.under_test)
        if isinstance(self.lower_level, HRL):
            self.lower_level.testing_sub_goal = self.testing_sub_goal

    def reset(self):
        self.control_algorithm.reset()
        self.high_level_agent.reset()
        self.__init__(self.control_algorithm, self.layer_id + 1, self.high_level_agent, **self.init_params)


import abc
from abc import ABC
from math import pi
from typing import Union

from environments import NavigationEnv, GoalReachingNavEnv, MapsIndex
from gym.spaces import Box, Discrete
import numpy as np

from hbrl.environments.environment import Environment


class TurtleEnv(NavigationEnv):
    def __init__(self, map_tag: MapsIndex, reset_anywhere=False):
        state_bounds = np.array([0, 0, 1, 1, 2 * pi])
        state_space = Box(low=-state_bounds, high=state_bounds)
        action_space = Box(-1, 1, (2,))

        super().__init__(map_tag, state_space, action_space, reset_anywhere)

    def reset(self):
        # TODO
        pass
    def step(self, action):
        # TODO
        pass

class GCEnv(Environment, ABC):

    def __init__(self, wrapped_environment, goal_space: Union[None, Box, Discrete]=None):
        self.wrapped_environment = wrapped_environment
        assert isinstance(self.wrapped_environment, Environment)
        self.goal_space = self.state_space if goal_space is None else goal_space
        assert isinstance(self.goal_space, (Box, Discrete))
        self.goal = None
        self.reset()

    def __getattr__(self, name):
        """Returns an attribute with ``name``, unless ``name`` starts with an underscore."""
        if name.startswith("_"):
            raise AttributeError(f"accessing private attribute '{name}' is prohibited")
        return getattr(self.wrapped_environment, name)

    @property
    def goal_size(self):
        return self.goal_space.shape[0]

    def sample_goal(self):
        return self.state_space.sample()

    def reset(self):
        self.wrapped_environment.reset()
        self.goal = self.sample_goal()
        return self.state, self.goal

    @abc.abstractmethod
    def reached(self, state, goal=None) -> bool:
        pass

    @abc.abstractmethod
    def reward(self, state, action, goal=None):
        pass

    def get_goal_from_state(self, state):
        """
        Apply a projection P: S -> G on the given state
        """
        return state

    def get_state_from_goal(self, goal):
        """
        Apply a \bar{P}:G -> S projection on the given state
        """
        return goal


class GoalReachingTurtleEnv(TurtleEnv, GoalReachingNavEnv):

    def __init__(self, map_tag: MapsIndex, reset_anywhere=False):
        wrapped_env = TurtleEnv(map_tag, reset_anywhere)
        GoalReachingNavEnv.__init__(self, wrapped_env)

    def reached(self):
        print("blabla")

    def get_goal_from_state(self, state):
        print("blabla")

    def get_state_from_goal(self, goal):
        print("blabla")

if __name__ == "__main__":
    environment = TurtleEnv(MapsIndex.EMPTY)
    gc_environment = GCNavigationEnv(TurtleEnv, MapsIndex.EMPTY)